import logging


# Create a passthrough handler to avoid browser-use
# to setup its own logging configuration
# https://github.com/browser-use/browser-use/blob/5e0af71af83e0da594044e23758c9564f91a185f/browser_use/logging_config.py#L71
# Lila uses loguru.
class PassHandler(logging.Handler):
    def emit(self, record):
        pass


# Add a passthrough handler to the root logger
logging.getLogger().addHandler(PassHandler())



import os
import re
import sys
import uuid
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import requests
from browser_use.agent.views import ActionResult
from browser_use.browser.context import BrowserContext
from browser_use.browser.views import (
    BrowserState,
)
from browser_use.dom.views import (
    DOMElementNode,
)
from jinja2 import Template
from loguru import logger

from lila.const import INCLUDE_ATTRIBUTES


@dataclass
class Pointer:
    x: int
    y: int


def parse_tags(value: str) -> List[str]:
    ret = []
    if value:
        tags = value.split(",")
        for tag in tags:
            tag = tag.strip()
            if not re.match(r"^[a-zA-Z0-9-_]+$", tag.strip()):
                raise ValueError(
                    f"Invalid tag '{tag}'. Tags must be alphanumeric (with optional hyphens or underscores)."
                )
            ret.append(tag)

    return ret


def get_vars(content: str) -> List[str]:
    # Searches for all ${VAR_NAME} and returns the list
    # of strings VAR_NAME
    regex = re.compile(r"\${(.*?)}")
    ret = regex.findall(content)
    return ret


def get_missing_vars(vars_list: List[str]) -> List[str]:
    # Returns a list of variables in vars_list that are not present
    # in the environment variables
    ret = []
    for var in vars_list:
        if os.environ.get(var) is None:
            ret.append(var)

    return ret


def get_vars_from_env(
    vars_list: List[str], fail_if_missing: bool = True
) -> Dict[str, str]:
    # Returns a dictionary with the values of the environment variables
    # specified in vars_list
    ret = {}
    for var in vars_list:
        if os.environ.get(var) is None:
            if fail_if_missing:
                raise RuntimeError(f"Environment variable {var} not found")
        else:
            ret[var] = os.environ[var]

    return ret


def replace_vars_with_placeholders(
    content: str, vars_list: List[str]
) -> Tuple[str, Dict]:
    # Replaces all ${VAR_NAME} with placeholder_{idx}
    mapping = {}
    uuid_str = str(uuid.uuid4())[:4]
    for idx, var in enumerate(set(vars_list)):
        placeholder = f"placeholder_{uuid_str}"
        content = content.replace(f"${{{var}}}", placeholder)
        mapping[placeholder] = var

    return content, mapping


def replace_placeholders_with_values(content: str, mapping: Dict[str, str]) -> str:
    # Replaces all placeholder_{idx} with the os env var value
    for placeholder, value in mapping.items():
        new_value = os.environ.get(value)
        if new_value is None:
            raise RuntimeError(f"Environment variable {value} not found")
        content = content.replace(placeholder, new_value)

    return content


def dump_browser_state(
    state: BrowserState, prev_interaction: Optional[Dict] = None
) -> Dict:
    ret = {
        "dom": state.element_tree.clickable_elements_to_string(
            include_attributes=INCLUDE_ATTRIBUTES
        ),
        "url": state.url,
        "title": state.title,
        "screenshot_b64": state.screenshot,
        "pixels_above": state.pixels_above,
        "pixels_below": state.pixels_below,
    }
    if prev_interaction:
        ret["prev_interaction"] = prev_interaction

    return ret


def send_state(
    idx: int,
    run_id: str,
    server_url: str,
    state: BrowserState,
    prev_interaction: Optional[Dict] = None,
) -> None:
    dumped_state = dump_browser_state(state, prev_interaction)
    ret = requests.post(
        f"{server_url}/api/v1/remote/runs/{run_id}/steps/{idx}/states",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['LILA_API_KEY']}",
        },
        json=dumped_state,
    )
    ret.raise_for_status()


def get_completion(idx: int, run_id: str, server_url: str) -> Dict:
    ret = requests.get(
        f"{server_url}/api/v1/remote/runs/{run_id}/steps/{idx}/completion",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['LILA_API_KEY']}",
        },
    )
    ret.raise_for_status()
    return ret.json()


### Custom implementation of key press over an element
### Based on the original input controller action
### https://github.com/browser-use/browser-use/blob/87f4d74e063f3c6236a53238c7178564158f3966/browser_use/controller/service.py#L227
@dataclass
class PressKeyWithFocus:
    index: int
    key: str


async def _press_key_element_node(
    context: BrowserContext, element_node: DOMElementNode, key: str
):
    try:
        # Highlight before typing
        if element_node.highlight_index is not None:
            await context._update_state(focus_element=element_node.highlight_index)

        page = await context.get_current_page()
        element_handle = await context.get_locate_element(element_node)

        if element_handle is None:
            raise Exception(f"Element: {repr(element_node)} not found")

        await element_handle.scroll_into_view_if_needed(timeout=2500)
        await element_handle.focus()
        await element_handle.press(key)
        await page.wait_for_load_state()

    except Exception as e:
        raise Exception(
            f"Failed to press key into element: {repr(element_node)}. Error: {str(e)}"
        )


async def press_key(params: PressKeyWithFocus, context: BrowserContext):
    session = await context.get_session()
    state = session.cached_state

    if params.index not in state.selector_map:
        raise Exception(
            f"Element index {params.index} does not exist - retry or use alternative actions"
        )

    element_node = state.selector_map[params.index]
    await _press_key_element_node(context, element_node, params.key)
    msg = f"⌨️  Pressed {params.key} into index {params.index}"
    logger.debug(msg)
    logger.debug(f"Element xpath: {element_node.xpath}")
    return ActionResult(extracted_content=msg, include_in_memory=True)


###


def setup_logging(debug: bool):
    # Cleanup existing config
    logger.remove()

    def formatter(record):
        ret = ""
        if debug:
            ret += (
                "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
                "<level>{level: <8}</level> | "
                "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
            )

        if "test_name" in record["extra"]:
            MAX_STEP_LENGTH = 20
            if "step" in record["extra"]:
                if len(record["extra"]["step"]) > MAX_STEP_LENGTH:
                    ret += f'<cyan>[{record["extra"]["test_name"]}| {record["extra"]["step"][:MAX_STEP_LENGTH]}...]</cyan> '
                else:
                    ret += f'<cyan>[{record["extra"]["test_name"]}| {record["extra"]["step"]}]</cyan> '
            else:
                ret += f'<cyan>[{record["extra"]["test_name"]}]</cyan> '

        ret += f'<level>{record["message"]}</level>\n'
        return ret

    logger.add(
        sink=sys.stderr,
        level="DEBUG" if debug else "INFO",
        format=formatter,
    )

    logger.add(
        sink=".lila.log",
        level="DEBUG",
        format="{time} {level} {message}",
        rotation="1MB",
    )


def pretty_step_formatter(step: Dict) -> str:
    action, content = list(step.items())[0]
    verifications = step.get("verify", [])
    if isinstance(verifications, str):
        verifications = [verifications]

    return f"{action}: {content} | {len(verifications)} verifications"


def render_template_to_file(template_path, output_file, data):
    try:
        # Read the template file
        with open(template_path, "r", encoding="utf-8") as f:
            template_content = f.read()

        # Create template object
        template = Template(template_content)

        template.globals["pretty_step"] = pretty_step_formatter

        # Render the template
        rendered_html = template.render(**data)

        # Write the rendered HTML to file
        with open(output_file, "w", encoding="utf-8") as f:
            f.write(rendered_html)

        return True

    except Exception as e:
        logger.error(f"Error rendering template: {str(e)}")
        return False



import asyncio
import json
import os
import tempfile
import traceback
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

import requests
import yaml
from browser_use import Browser, BrowserConfig
from browser_use.browser.context import BrowserContext, BrowserContextConfig
from browser_use.controller.service import Controller
from browser_use.controller.views import (
    ClickElementAction,
    InputTextAction,
    ScrollAction,
)
from loguru import logger

from lila.config import Config
from lila.const import MAX_ATTEMPTS, MAX_LOGS_DISPLAY
from lila.utils import (
    PressKeyWithFocus,
    dump_browser_state,
    get_completion,
    press_key,
    pretty_step_formatter,
    render_template_to_file,
    send_state,
)


@dataclass
class ReportLog:
    log: str
    screenshot_b64: str


class FailedStepError(RuntimeError):
    pass


def report_test_status(run_id: str, payload: Dict, server_url: str) -> None:
    ret = requests.patch(
        f"{server_url}/api/v1/remote/runs/{run_id}",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['LILA_API_KEY']}",
        },
        json=payload,
    )
    ret.raise_for_status()
    logger.debug(f"Successfully reported test status for run {run_id}")


def report_step_result(
    run_id: str, idx: int, result: str, msg: str, server_url: str
) -> None:
    ret = requests.put(
        f"{server_url}/api/v1/remote/runs/{run_id}/steps/{idx}/result",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['LILA_API_KEY']}",
        },
        json={"result": result, "msg": msg},
    )
    ret.raise_for_status()
    logger.debug(f"Successfully reported step result for run {run_id}")


def send_step_screenshot(
    run_id: str, idx: int, screenshot_b64: str, server_url: str
) -> None:
    ret = requests.post(
        f"{server_url}/api/v1/remote/runs/{run_id}/steps/{idx}/screenshots",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['LILA_API_KEY']}",
        },
        json={"screenshot_b64": screenshot_b64},
    )
    ret.raise_for_status()
    logger.debug(f"Successfully sent screenshot for run {run_id}")


def send_step_log(run_id: str, idx: int, level: str, msg: str, server_url: str) -> None:
    ret = requests.post(
        f"{server_url}/api/v1/remote/runs/{run_id}/steps/{idx}/logs",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['LILA_API_KEY']}",
        },
        json={"level": level, "msg": msg},
    )
    ret.raise_for_status()
    logger.debug(f"Successfully sent log for run {run_id}")


@dataclass
class StepResult:
    success: bool
    msg: str


controller = Controller()


@dataclass
class TestCase:
    name: str
    steps: List[Dict[str, str]]

    tags: List[str] = field(default_factory=list)
    raw_content: str = ""

    status: str = "pending"
    steps_results: List[StepResult] = field(default_factory=list)

    logs: List[Dict[str, Any]] = field(default_factory=list)

    duration: float = 0.0

    _should_stop: bool = False

    @classmethod
    def from_yaml(cls, name: str, content: str):
        data = yaml.safe_load(content)
        logger.debug(f"Loaded test case: {name}")
        return cls(
            name=name,
            steps=[step for step in data["steps"]],
            tags=data.get("tags", []),
            status="pending",
            raw_content=content,
        )

    def dump_report(self, output_dir: str, report: Dict[int, List[ReportLog]]) -> None:
        # Writes a markdown report for the test case
        template_data = {
            "name": self.name,
            "steps": self.steps,
            "report": report,
            "now": datetime.utcnow(),
        }
        template_path = Path(__file__).parent / "assets" / "report.html"
        output = Path(output_dir) / f"{self.name}.html"
        render_template_to_file(template_path, output, template_data)
        logger.debug(f"Report for {self.name} saved at {output}")

    def _update_state(self, run_id: str, server_url: str):
        ret = requests.get(
            f"{server_url}/api/v1/remote/runs/{run_id}/status",
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
                "Authorization": f"Bearer {os.environ['LILA_API_KEY']}",
            },
        )
        ret.raise_for_status()
        data = ret.json()
        logger.debug(f"Succesfully fetched state for run {run_id}")
        self.steps_results = [
            StepResult(success=step["success"], msg=step["msg"])
            for step in data["step_results"]
        ]
        self.logs = [
            {"level": log["level"], "msg": log["msg"]} for log in data["logs"]
        ][-MAX_LOGS_DISPLAY:]
        self.status = data.get("conclusion", data["run_status"])
        logger.debug("Updated update queue for new state to render")

    async def handle_verifications(
        self,
        idx: int,
        step_content: str,
        context: BrowserContext,
        run_id: str,
        server_url: str,
        report_logs: List[ReportLog],
        dry_run: bool = False,
    ) -> None:
        state = await context.get_state()
        dumped_state = dump_browser_state(state)
        ret = requests.post(
            f"{server_url}/api/v1/remote/runs/{run_id}/steps/{idx}/verifications",
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
                "Authorization": f"Bearer {os.environ['LILA_API_KEY']}",
            },
            json=dumped_state,
        )
        ret.raise_for_status()
        logger.debug(
            f"Successfully sent browser state for verification for step {idx } run {run_id}"
        )

        data = ret.json()

        if not data["verifications"]:
            logger.debug(f"No verifications found for step {idx}")
            return

        for verification in data["verifications"]:
            if verification["status"] == "failure":
                logger.debug(
                    f"Verification failed for step {pretty_step_formatter(self.steps[idx])}: {verification['log']}"
                )
                report_logs.append(
                    ReportLog(
                        log=f"Verification failed: {verification['log']}",
                        screenshot_b64=await context.take_screenshot(),
                    )
                )
                raise FailedStepError(f"Failed to execute step: {verification['log']}")

            logger.debug(f"Verification passed for step {idx}: {verification['log']}")
            report_logs.append(
                ReportLog(
                    log=f"Verification passed: {verification['log']}",
                    screenshot_b64=await context.take_screenshot(),
                )
            )

    async def _handle_complex_step(
        self,
        idx: int,
        context: BrowserContext,
        run_id: str,
        server_url: str,
        report_logs: List[ReportLog],
        dry_run: bool = False,
    ) -> None:
        done = False
        prev_interaction = None
        scrolling = False
        attempts = 0
        while not done:
            action_result = None
            attempts += 1
            if attempts > MAX_ATTEMPTS:
                raise FailedStepError("Step took too long to complete")

            state = await context.get_state()
            send_state(idx, run_id, server_url, state, prev_interaction)
            logger.debug(f"Sent state for step {idx}")

            data = get_completion(idx, run_id, server_url)["completion"]

            report_logs.append(
                ReportLog(
                    log=data["log"],
                    screenshot_b64=await context.take_screenshot(),
                )
            )

            logger.debug(f"Got completion data for step {idx}")

            logger.info(data["log"])
            if "action" in data:
                logger.debug(f"Completion requires action: {data['action']}")
                if data["action"] == "click":
                    fn = controller.registry.registry.actions["click_element"]
                    action_result = await fn.function(
                        ClickElementAction(index=int(data["element"])), context
                    )
                    # Hack. There is a bug in browser use that reports a click
                    # as a failure when the context is destroyed due to a navigation
                    # event. This is a workaround to handle that.
                    if (
                        action_result.error
                        and "Execution context was destroyed" in action_result.error
                    ):
                        action_result.error = None

                    if action_result.error:
                        logger.error(f"Click failed: {action_result.error}")
                    else:
                        logger.info("Click performed successfully")

                    prev_interaction = {
                        "status": "failure" if action_result.error else "success",
                        "msg": action_result.error
                        or "Successfully clicked requested element",
                    }
                elif data["action"] == "input":
                    fn = controller.registry.registry.actions["input_text"]
                    action_result = await fn.function(
                        InputTextAction(index=int(data["element"]), text=data["text"]),
                        context,
                    )

                    if action_result.error:
                        logger.error(f"Input failed: {action_result.error}")
                    else:
                        logger.info("Input performed successfully")

                    prev_interaction = {
                        "status": "failure" if action_result.error else "success",
                        "msg": action_result.error
                        or "Successfully inputted text into requested element",
                    }
                elif data["action"] == "press-enter":
                    action_result = await press_key(
                        PressKeyWithFocus(index=int(data["element"]), key="Enter"),
                        context,
                    )
                    logger.debug("Pressed Enter key")
                    if action_result.error:
                        logger.error(f"Press Enter failed: {action_result.error}")
                    else:
                        logger.info("Press Enter performed successfully")

                    prev_interaction = {
                        "status": "failure" if action_result.error else "success",
                        "msg": action_result.error or "Successfully pressed Enter key",
                    }
                elif data["action"] == "native-dropdown-open":
                    fn = controller.registry.registry.actions["get_dropdown_options"]
                    action_result = await fn.function(
                        index=int(data["element"]),
                        browser=context,
                    )

                    if action_result.error:
                        logger.error(f"Failed to open dropdown: {action_result.error}")
                    else:
                        logger.info("Opened dropdown and got options")
                    prev_interaction = {
                        "status": "failure" if action_result.error else "success",
                        "msg": action_result.error
                        or f"Successfully opened dropdown and got options:\n{action_result.extracted_content}",
                    }
                    logger.debug("Opened dropdown and got options")
                elif data["action"] == "native-dropdown-select":
                    fn = controller.registry.registry.actions["select_dropdown_option"]
                    action_result = await fn.function(
                        index=int(data["element"]),
                        text=data["text"],
                        browser=context,
                    )

                    if action_result.error:
                        logger.error(
                            f"Failed to select option from dropdown: {action_result.error}"
                        )
                    else:
                        logger.info("Selected option from dropdown")

                    prev_interaction = {
                        "status": "failure" if action_result.error else "success",
                        "msg": action_result.error
                        or "Successfully selected option from dropdown",
                    }
                    logger.debug("Opened dropdown and got options")
                else:
                    raise RuntimeError(f"Unknown action {data['action']}")

                await context._wait_for_page_and_frames_load()
                report_logs.append(
                    ReportLog(
                        log=action_result.error or "Action performed successfully",
                        screenshot_b64=await context.take_screenshot(),
                    )
                )

            if data["status"] == "complete" and (
                action_result is None or not action_result.error
            ):
                done = True
                logger.debug("Completion reports step completed.")
                report_logs.append(
                    ReportLog(
                        log="Step completed successfully",
                        screenshot_b64=await context.take_screenshot(),
                    )
                )
            if data["status"] == "wip":
                logger.debug("Completion reports step still in progress.")
            elif data["status"] == "not-found":
                logger.debug("Completion reports element not found. Running scrolling.")
                if scrolling or state.pixels_above == 0:
                    if state.pixels_below == 0:
                        logger.debug("No more elements to uncover. Failing step.")
                        raise FailedStepError(f"Failed to execute step: {data['log']}")

                    fn = controller.registry.registry.actions["scroll_down"]
                    await fn.function(ScrollAction(), context)
                    logger.debug("Scrolled one page down")
                    prev_interaction = {
                        "status": "success",
                        "msg": "Scrolled down one page to uncover more elements.",
                    }
                    report_logs.append(
                        ReportLog(
                            log="Scrolled down one page to uncover more elements.",
                            screenshot_b64=await context.take_screenshot(),
                        )
                    )
                else:
                    fn = controller.registry.registry.actions["scroll_up"]
                    await fn.function(ScrollAction(state.pixels_above), context)
                    logger.debug("Moved all the way up")
                    prev_interaction = {
                        "status": "success",
                        "msg": "Scrolled to the top of the page to uncover more elements.",
                    }
                    report_logs.append(
                        ReportLog(
                            log="Scrolled to the top of the page to uncover more elements.",
                            screenshot_b64=await context.take_screenshot(),
                        )
                    )

                scrolling = True
                logger.debug("Activated scrolling mode")

                await context._wait_for_page_and_frames_load()
            elif data["status"] == "failure":
                logger.debug(f"Completion reports step failed: {data['log']}")
                report_logs.append(
                    ReportLog(
                        log="Failed to execute step: {data['log']}",
                        screenshot_b64=await context.take_screenshot(),
                    )
                )
                raise FailedStepError(f"Failed to execute step: {data['log']}")

    async def handle_step(
        self,
        idx: int,
        step_type: str,
        step_content: str,
        verifications: List[str],
        context: BrowserContext,
        run_id: str,
        server_url: str,
        report_logs: List[ReportLog],
        dry_run: bool = False,
    ) -> None:
        page = await context.get_current_page()
        if step_type == "goto":
            await page.goto(step_content)
            await context._wait_for_page_and_frames_load()
            logger.info("Navigation completed successfully")
            report_logs.append(
                ReportLog(
                    log="Navigation completed successfully",
                    screenshot_b64=await context.take_screenshot(),
                )
            )
        else:
            await self._handle_complex_step(
                idx, context, run_id, server_url, report_logs, dry_run
            )

        if verifications:
            logger.info("Running step verifications")
            await self.handle_verifications(
                idx, step_content, context, run_id, server_url, report_logs, dry_run
            )
            logger.debug("Verifications completed successfully")
        else:
            logger.info("No verifications for this step")

    def start(self, server_url: str, batch_id: str) -> str:
        ret = requests.post(
            f"{server_url}/api/v1/remote/runs",
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
                "Authorization": f"Bearer {os.environ['LILA_API_KEY']}",
            },
            json={
                "name": self.name,
                "content": self.raw_content,
                "batch_id": batch_id,
            },
        )
        ret.raise_for_status()
        if ret.status_code != 201:
            raise RuntimeError(f"Failed to start run: {ret.json()}")

        data = ret.json()
        logger.info(f"Successfully started running test {self.name} [{data['run_id']}]")
        return data["run_id"]

    @staticmethod
    def initialize_browser_context(
        config: Config, browser_state: Optional[str] = None
    ) -> BrowserContext:
        storage_state = None
        if browser_state:
            logger.info(f"Loading browser state from {browser_state}")
            with open(browser_state, "r") as f:
                storage_state = json.load(f)

        browser_config = BrowserConfig(headless=config.browser.headless)
        browser = Browser(config=browser_config)

        cookies_file = None
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".json") as f:
            if storage_state and "cookies" in storage_state:
                json.dump(storage_state["cookies"], f)
                cookies_file = f.name

            # Ref https://docs.browser-use.com/customize/browser-settings#context-configuration
            context_config = BrowserContextConfig(
                wait_for_network_idle_page_load_time=3,
                browser_window_size={
                    "width": config.browser.width,
                    "height": config.browser.height,
                },
                # pass -1 for highlighting all elements in the page
                viewport_expansion=0,
                cookies_file=cookies_file,
            )

            return BrowserContext(browser=browser, config=context_config)

    @staticmethod
    async def teardown(context: BrowserContext, name: str, config: Config) -> None:
        path = Path(config.runtime.output_dir) / f"{name}.json"
        os.makedirs(path.parent, exist_ok=True)
        await context.session.context.storage_state(path=path)
        logger.debug(f"Browser state saved for {name} at {path}")
        await context.close()
        await context.browser.close()

    async def run(
        self,
        run_id: str,
        config: Config,
        browser_state: Optional[str],
        name: str,
        dry_run: bool = False,
    ) -> bool:
        with logger.contextualize(test_name=name):
            server_url = config.runtime.server_url
            report_test_status(
                run_id, {"status": "running" if not dry_run else "dry-run"}, server_url
            )
            logger.debug(f"Test {run_id} started running")

            context = self.initialize_browser_context(config, browser_state)
            logger.debug(f"Browser context initialized for {run_id}: {context}")

            success = True
            report: Dict[int, List[ReportLog]] = {}
            for idx, step in enumerate(self.steps):
                report[idx] = []
                step_type, step_content = [(k, v) for k, v in step.items()][0]
                with logger.contextualize(step=f"{step_type} {step_content}"):
                    verifications: str | List[str] = step.get("verify", [])
                    if isinstance(verifications, str):
                        verifications_list: List[str] = [verifications]
                    else:
                        verifications_list = verifications

                    logger.debug(
                        f"Running step {idx}: {step_type} {step_content} [{len(verifications_list)} verifications]"
                    )
                    try:
                        await self.handle_step(
                            idx,
                            step_type,
                            step_content,
                            verifications_list,
                            context,
                            run_id,
                            server_url,
                            report[idx],
                            dry_run=dry_run,
                        )
                        logger.success("Step completed successfully")
                    except FailedStepError as e:
                        success = False
                        logger.error(f"Failed to execute step: {str(e)}")
                        break
                    except Exception as e:
                        logger.exception(f"Unexpected error: {e}")
                        await self.teardown(context, name, config)
                        raise

            if not dry_run:
                report_test_status(
                    run_id,
                    {
                        "status": "finished",
                        "conclusion": "success" if success else "failure",
                    },
                    server_url,
                )

            await self.teardown(context, name, config)
            self.dump_report(config.runtime.output_dir, report)
            report_path = Path(config.runtime.output_dir) / f"{name}.html"
            logger.info(f"Report saved at {report_path}")
            return success


def collect_test_cases(test_files: List[str], tags: List[str], exclude_tags: List[str]):
    testcases = []
    for path in test_files:
        with open(path, "r") as f:
            content = f.read()
            # Remove extension for filename
            name = os.path.splitext(path)[0]
            test = TestCase.from_yaml(name, content)

        if tags:
            if not set(tags).intersection(test.tags):
                continue

            if exclude_tags:
                if set(exclude_tags).intersection(test.tags):
                    continue

        testcases.append(test)

    return testcases


class TestRunner:
    def __init__(self, testcases: List[TestCase]):
        self.testcases = testcases

    def run_tests(
        self, config: Config, browser_state: Optional[str], batch_id: str, dry_run: bool
    ) -> bool:
        future_to_test = {}

        with ThreadPoolExecutor(
            max_workers=config.runtime.concurrent_workers
        ) as executor:
            for idx, testcase in enumerate(self.testcases):
                with logger.contextualize(test_name=testcase.name):
                    run_id = testcase.start(config.runtime.server_url, batch_id)

                    # For debuggin purposes
                    def run_wrapper(*args, **kwargs):
                        try:
                            return asyncio.run(testcase.run(*args, **kwargs))
                        except Exception:
                            print(traceback.format_exc())
                            raise

                    # Submit all tests
                    key = executor.submit(
                        run_wrapper,
                        run_id,
                        config,
                        browser_state,
                        name=testcase.name,
                        dry_run=True,
                    )
                    future_to_test[key] = testcase

            for future in as_completed(future_to_test.keys()):
                result = future.result()
                testcase = future_to_test[future]

                testcase.status = "success" if result else "failure"

        # Show summary and failed test details
        total = len(self.testcases)
        passed = sum(1 for t in self.testcases if t.status == "success")
        failed = sum(1 for t in self.testcases if t.status == "failure")

        if failed:
            logger.error(
                f"Test Report - Passed: {passed}, Failed: {failed}, Total: {total}"
            )
            return False
        else:
            logger.success(
                f"Test Report - Passed: {passed}, Failed: {failed}, Total: {total}"
            )
            return True



BASE_URL = "https://app.lila.dev"
WIDTH = 1536
HEIGHT = 1152

MAX_LOGS_DISPLAY = 3

API_KEY_URL = "https://app.lila.dev/manage-api-keys"

INCLUDE_ATTRIBUTES = [
    "title",
    "type",
    "name",
    "role",
    "tabindex",
    "aria-label",
    "placeholder",
    "value",
    "alt",
    "aria-expanded",
]

MAX_ATTEMPTS = 30



import json
import os
import pathlib
import shutil
import sys
import uuid
from pathlib import Path
from typing import Dict, List, Optional

import click
import requests
import yaml
from loguru import logger

from lila.config import Config
from lila.const import API_KEY_URL
from lila.runner import TestRunner, collect_test_cases
from lila.utils import parse_tags, setup_logging


def validate_content(content: str, server_url: str) -> None:
    try:
        yaml.safe_load(content)
    except yaml.YAMLError as e:
        raise ValueError(f"Provided content is not a valid YAML: {e}")

    ret = requests.post(
        f"{server_url}/api/v1/testcase-validations",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['LILA_API_KEY']}",
        },
        json={"content": content},
    )
    raise_for_status(ret)
    if ret.status_code == 200:
        data = ret.json()
        if data["valid"]:
            return

        raise ValueError(
            f"Provided content is not a valid Lila test case: {data['message']}"
        )


def find_parsing_errors(test_paths: List[str], server_url: str) -> Dict[str, str]:
    invalid_tests = {}
    for path in test_paths:
        with open(path, "r") as f:
            content = f.read()
            try:
                validate_content(content, server_url)
            except ValueError as e:
                invalid_tests[path] = str(e)

    return invalid_tests


def raise_for_status(response: requests.Response):
    try:
        response.raise_for_status()
    except requests.RequestException as e:
        try:
            data = response.json()
            raise RuntimeError(f"Error: {data}") from e
        except json.JSONDecodeError:
            raise RuntimeError(f"Error: {response.text}") from e


@click.group()
def cli():
    """Lila CLI tool."""
    pass


def collect(path) -> List[str]:
    """
    Find all YAML files in a given path. If path is a file, return it if it's a YAML file.
    If path is a directory, recursively search for all YAML files within it.

    Args:
        path (str): Path to file or directory

    Returns:
        list: List of paths to YAML files found
    """
    yaml_extensions = (".yaml", ".yml")
    result = []

    # Convert path to Path object for easier handling
    path_obj = pathlib.Path(path)

    # If path is a file, check if it's a YAML file
    if path_obj.is_file():
        if path_obj.suffix.lower() in yaml_extensions:
            return [str(path_obj)]
        return []

    # If path is a directory, walk through it recursively
    if path_obj.is_dir():
        for root, _, files in os.walk(path):
            for file in files:
                file_path = pathlib.Path(root) / file
                if file_path.suffix.lower() in yaml_extensions:
                    result.append(str(file_path))

    return sorted(result)  # Sort for consistent output


def _get_config(config_file: Optional[str]) -> Config:
    if not config_file:
        if os.path.exists("lila.toml"):
            config_file = "lila.toml"
        else:
            return Config.default()
    elif not os.path.exists(config_file):
        raise FileNotFoundError(f"Config file not found: {config_file}")

    return Config.from_toml_file(config_file)


@cli.command()
@click.argument("path", type=str, required=True)
@click.option(
    "--tags",
    type=str,
    help="Comma-separated list of tags",
    required=False,
)
@click.option("--exclude-tags", type=str, help="Exclude tests by tags")
@click.option("--config", type=str, help="Path to the Lila config file", required=False)
@click.option(
    "--browser-state", type=str, help="Path to the browser state file", required=False
)
@click.option(
    "--output-dir", type=str, help="Override config output directory", required=False
)
@click.option(
    "--debug",
    is_flag=True,
    help="Enable debug mode",
    default=False,
)
@click.option(
    "--headless",
    is_flag=True,
    help="Run tests in headless mode",
    default=False,
)
def run(
    path: str,
    tags: str,
    exclude_tags: str,
    config: Optional[str],
    browser_state: Optional[str],
    output_dir: Optional[str],
    debug: bool = False,
    headless: bool = False,
):
    """Run a Lila test suite."""
    setup_logging(debug=debug)
    # This were previously run flags,
    # will add the later once webapp
    # is more stable.
    dry_run = True
    batch_id = None

    if "LILA_API_KEY" not in os.environ:
        logger.error(
            "Please set the LILA_API_KEY environment variable. You can find it in the Lila app: %s",
            API_KEY_URL,
        )
        return

    try:
        config_obj = _get_config(config)
    except FileNotFoundError as e:
        logger.error(str(e))

    if output_dir:
        config_obj.runtime.output_dir = output_dir

    config_obj.browser.headless = headless

    tag_list = []
    if tags:
        try:
            tag_list = parse_tags(tags)
        except ValueError as e:
            logger.error(str(e))
            return

    exclude_tag_list = []
    if exclude_tags:
        try:
            exclude_tag_list = parse_tags(exclude_tags)
        except ValueError as e:
            logger.error(str(e))
            return

    # If intersection of tags and exclude_tags is not empty, raise an error
    if set(tag_list) & set(exclude_tag_list):
        logger.error(
            "Tags and exclude-tags cannot have common elements: %s and %s",
            tag_list,
            exclude_tag_list,
        )
        return

    if browser_state and not os.path.exists(browser_state):
        logger.error("Browser state file not found: %s", browser_state)
        return

    test_files = collect(path)
    if not test_files:
        logger.error("No YAML files found in the provided path: %s", path)
        return

    invalid_files = find_parsing_errors(test_files, config_obj.runtime.server_url)
    if invalid_files:
        logger.error("Parsing errors found")
        for path, error in invalid_files.items():
            logger.error("File %s: %s", path, error)
        return

    testcases = collect_test_cases(test_files, tag_list, exclude_tag_list)

    if not testcases:
        logger.error(
            "No test cases found with the provided path and the provided params"
        )
        return

    if not batch_id:
        batch_id = str(uuid.uuid4())

    runner = TestRunner(testcases)
    if not dry_run:
        logger.info("Tracking URL: %s/runs/%s", config_obj.runtime.server_url, batch_id)
    success = runner.run_tests(config_obj, browser_state, batch_id, dry_run)
    if not success:
        sys.exit(1)


@cli.command()
def init():
    """Initialize a Lila testing template."""
    if os.path.exists("lila.toml"):
        logger.error(
            "Config file lila.toml already exists, it seems like the application is already initialized."
        )
        return

    # Create a lila.toml file
    config_path = Path(__file__).parent / "assets" / "lila.toml"
    shutil.copy(config_path, "lila.toml")
    logger.info("Config file created: lila.toml")

    # Create a lila directory
    os.makedirs("lila-output", exist_ok=True)
    logger.info("Output directory created for artifacts: lila-output/")

    # Create an example test case
    example_path = Path(__file__).parent.parent / "tests" / "google-maps.yaml"
    shutil.copy(example_path, "demo.yaml")
    logger.info("Example test case created: demo.yaml")
    logger.success("All set! Run your first test: lila run demo.yaml")


@click.option("--config", type=str, help="Path to the Lila config file", required=False)
@cli.command()
def check(
    config: Optional[str],
) -> None:
    """Check if Lila is properly set up and ready to run tests."""
    setup_logging(debug=False)
    if "LILA_API_KEY" not in os.environ:
        logger.error(
            f"Please set the LILA_API_KEY environment variable. You can find it in the Lila app: {API_KEY_URL}"
        )
        return

    config_obj = _get_config(config)
    server_url = config_obj.runtime.server_url
    ret = requests.get(
        f"{server_url}/api/v1/readiness",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['LILA_API_KEY']}",
        },
    )
    if ret.status_code == 401:
        logger.error("Invalid LILA_API_KEY. Please set the correct API key.")
        sys.exit(1)
        return

    if ret.status_code != 200:
        logger.error(f"Lila is not ready to run tests [code: {ret.status_code}]")
        sys.exit(1)
        return

    logger.success("Lila is properly set up and ready to run tests.")
